
/*
    Here are some definitions to make the parallel ports easier

    Use the value of ARCH to specify the TYPE of parallel machine
 */

#include "tools.h"

#ifndef __commh
#define __commh

#if defined(tc2000) || defined(encore)
#define SHARED_MEMORY
#endif
#if defined(intelnx) || defined(p4) || defined(pvm) || defined(picl) || \
    defined(eui)
#define DISTRIBUTED_MEMORY
#endif

#ifdef DEBUG_TRACEBACK
/* Redefine these to include to processor number */
#undef DEBUGTBF
#undef DEBUGTBCF
#define DEBUGTBF(line,file) \
    fprintf( stderr, "[%d] Line %d in %s\n", PImytid, line, file )
#define DEBUGTBCF(line,file,c) \
    fprintf( stderr, "[%d] Line %d in %s: %s\n", PImytid, line, file, c )
#endif

#include "comm/nbr.h"
#include "comm/procset.h"

extern void PIBroadcastArgs();
extern int  PIcall();
extern void PImerge();
extern void PIArgsToResources();
extern void PIGetResourceLimits();
extern void PIChangeResourceLimits();
extern void PISetOption();
extern int  PIGetNbrsHC(), PIGetNbrsMSH();

/* Here we define some parameter checking routines. */
#ifdef DEBUG_ALL
#define COMMCHECKSEND(typ,len,to) {\
    if ((typ) < -1) SETERRC(1,"Invalid type in send");\
    if ((len) < 0)  SETERRC(1,"Invalid (negative) length in send");\
    if ((to) < 0 || (to)>=NUMNODES) SETERRC(1,"Invalid destination");}
/* maxlen field of CHECKRECV isn't used yet */
#define COMMCHECKRECV(typ,maxlen) {\
    if ((typ) < -1)   SETERRC(1,"Invalid type in recv");\
    if ((maxlen) < 0)  SETERRC(1,"Invalid (negative) length in recv");}
#else
#define COMMCHECKSEND(type,len,to)
#define COMMCHECKRECV(type,len)
#endif
#include "comm/logging.h"

/*
    Eventually, we want to cache the node # and number of nodes in
    these variables so that the cost of logging and other checks
    is as low as possible.  Since these need to be set, for those
    systems that can be used WITHOUT PIcall, we need to provide an
    "init" routine that must be called.  Just to make these slightly
    safer, they will be initialized to -1 so that an "unset" condition
    is easy to establish.
*/
#if defined(pvm) || defined(picl)
extern int __MYPROCID, __NUMNODES;
#endif

/*D
    MessagePassing -  Common message passing macros.  

    Description:
    These convert into the popular forms of Intel NX, P4, and others 
    (eventually PICL, PVM, and others as contributed)
 
    These include send, receive, and global operations.  To encourage
    truely portable use of global operations, they all support processor
    subsets; if the native routine can't do that, then an
    implementation-specific version must be provided.  They also support
    asynchronous and "force" operations.

    There are many possible combinations.  They are aranged as

    SEND/RECV SYNC/ASYNC <null>/NOMEM <null>/FORCE

    <null> means that the item may be omitted.  The meanings are

    SEND/RECV - send or receive.

    SYNC/ASYNC - synchronous or asynchronous.  When the synchronous version
    returns, the buffer may be reused.  On a receive, the buffer contains
    a message.  When the asynchronous version returns, the "id" argument
    is a context that must be waited on (see SEND/RECV WAIT) before
    the buffer may be used.  The "id" value returned by these is > 0 on
    success.

    NOMEM - In order to allow the most flexible use of message passing
    libraries, this interface allows message buffers to be specially 
    allocated (see MSGALLOC SEND/RECV).  If the NOMEM version of a 
    routine is selected, the buffer was NOT allocated with the MSGALLOC
    macros.  The buffer MAY be used, OR a local buffer may be created 
    and freed.

    FORCE - (applies only to SEND) Send without any handshaking with the
    receiver.  This amy be used when it is certain that the receiver
    is already waiting to receive a message.  This should be used with
    great care; if the receiver is not ready for the message, the message
    may be discarded (the behavior is undefined).

$    SENDASYNC(type,buffer,length,to,datatype,id)
$    RECVASYNC(type,buffer,maxlength,datatype,id)
$    SENDSYNC(type,buffer,length,to,datatype)
$    RECVSYNC(type,buffer,maxlength,datatype)

    xxxxxxxxNOMEM - same as SEND/RECV ASYNC/SYNC, except the buffers
    have not been allocated with the MSGALLOC SEND/RECV macrocs.
    xxxxxxxxxxxxxFORCE - same as SEND/RECV ASYNC/SYNC <null>/NOMEM, except
    the destination must be ready to receive the message.

$    SENDWAIT(same as sendasync)
$    RECVWAIT(same as recvasync)
$    SENDWAITNOMEM ... ect

    In addition, a receive that handles an arbitrary length message is 
    provided

    RECVSYNCUNSZ(type,bufp,size,datatype)

    returns a pointer to a buffer "bufp" that contains a message of
    size "size".  The buffer may be freed with the usual macro.

    Message inquiry functions    
$    RECVLEN()      - length in bytes of the most recently received message
$    RECVFROM()     - id of the sender of the most recently received message

    A simplified mechanism for probing the receive queue is provided.
    A Probe determines if a message of a given type is available.
    If the probe is successful, RECVFROM may be used to determine
    who the message is from (the value of RECVLEN is implementation
    dependent).
$    SYNCPROBE(type)
$    ASYNCPROBE(type)

    Because of some indetminincy in some implementations, if you want
    to receive the message that you have just probed, use
    RECVPROBED(type,buffer,maxlength,datatype)

$    GxSUM(val,n,work,procset)
$    GxMAX(val,n,work,procset)
$    GxMIN(val,n,work,procset)

    (There should be more; they can be provided as requested.  There is
    also a function for providing an arbitrary, user-defined combination
    operation).

$    GCOL(lbuf,lsize,gbuf,gsiz,glen,procset,datatype) - global collection
$    GCOLX(lbuf,gsizes,gbuf,procset,datatype)         - global collection in 
                                              node number order with
					      given sizes

$    GSCATTER(buf,siz,issrc,procset,datatype)- buf is sent from src to everyone
                                              else in the procset
$    GSCATTERSRC(buf,size,src,procset,datatype)- like GSCATTER, except from
                                              a particular source

    GTOKEN(procset,i) - returns 1 if the processor is the i'th member of
    the procset and 0 otherwise.  In addition, a partial ordering is
    enforced: processor i can not proceed until processor (i-1) sees the
    value i.  This should ONLY be used in a loop like

        for (i=0;i<=np;i++) {if (gtoken(procset,i)) <do-something>;}

    where np is the number of processors in the procset.

    The "G" routines are global routines; if "procset" is non-null, they
    are global over a subset of processors.

    System inquiry functions
$    NUMNODES   - Number of processors
$    MYPROCID   - id (in [0:(NUMNODES-1)] of this processor
$    MSGTYPERANGE(&low,&high) - sets low to the lowest permissibly value,
                               high to the highest.  This is implementation
                               dependent, and not all systems make this
                               information available.
$    MSGSIZES(&min,&max)      - sets min to the minimum message size and
                               max to the maximum message size, in bytes.

    The following are not yet implemented
    
$    MSGMODELTIME(&latency,&timeperbyte,&hoptime) -
    		Models the time to send data as
		(latency + n_bytes * timeperword + n_hops * hoptime),
		where n_hops is the "distance" and n_bytes is the number
		of bytes being sent.  This information may not be available;
		the fall-back values are (0,1,0).  Values are in ?
		(we will decide after designing the pipelined broadcasting
		code).
$    MSGPACKETSIZE - returns the natural size in bytes of a message packet.
		    Value represents a natural size for a message
                    (often, the actual size of a message).  If < 0, there
                    is no (known) natural size.
$    MSGDISTANCE(from,to) - yields the number of hops or connections
                           in sending a message from processor "from" to
                           processor "to".  Since this is not aways known,
                           a value of "1" is the default except for
                           from == to.
                           
    With these, we could add an ADDITIONAL send/recv option: pipelined
    (not clear we want this for async routines)
 D*/

#if defined(p4)
#include "comm/commp4.h"
#elif defined(picl)
#include "comm/commpicl.h"
#elif defined(pvm)
#include "comm/commpvm.h"
#elif defined(intelnx) || defined(meiko)
#include "comm/commnx.h"
#elif defined(cm5)
#include "comm/commcm5.h"
#elif defined(eui)
#include "comm/commeui.h"
#elif defined(ncube)
#include "comm/commn3.h"
#else
#include "comm/commdef.h"
#endif

/* Some standard defines for common cases */
#ifdef NO_ASYNC_SEND
typedef int ASYNCSendId_t;
#define SENDASYNC(type,buffer,length,to,datatype,id) \
        SENDSYNC(type,buffer,length,to,datatype)
#define SENDASYNCNOMEM(type,buffer,length,to,datatype,id) \
        SENDASYNC(type,buffer,length,to,datatype,id)
#define SENDWAIT(type,buffer,length,to,datatype,id) 
#define SENDWAITNOMEM(type,buffer,length,to,datatype,id)  
#define SENDCANCEL(id)
#endif

#ifdef NO_ASYNC_RECV
typedef int ASYNCRecvId_t;
#define RECVASYNC(type,buffer,length,datatype,id) {id=1;}   
#define RECVASYNCNOMEM(type,buffer,length,datatype,id) {id=1;}
#define RECVWAIT(type,buffer,length,datatype,id) \
        RECVSYNC(type,buffer,length,datatype)
#define RECVWAITNOMEM(type,buffer,length,datatype,id)     \
        RECVSYNCNOMEM(type,buffer,length,datatype)
#define RECVCANCEL(id)
#endif

#if defined(NO_ASYNC_RECV) || defined(NO_ASYNC_SEND)
#define ASYNCDONE(id) 1
#endif

#ifdef NO_FORCE
#define SENDASYNCFORCE(type,buffer,length,to,datatype,id) \
        SENDSYNC(type,buffer,length,to,datatype)
#define RECVASYNCFORCE(type,buffer,length,datatype,id) \
        RECVASYNC(type,buffer,length,datatype,id) 
#define SENDSYNCFORCE(type,buffer,length,to,datatype) \
        SENDSYNC(type,buffer,length,to,datatype)
#define RECVSYNCFORCE(type,buffer,length,datatype) \
        RECVSYNC(type,buffer,length,datatype)
#define SENDASYNCNOMEMFORCE(type,buffer,length,to,datatype,id) \
        SENDASYNCNOMEM(type,buffer,length,to,datatype,id)
#define RECVASYNCNOMEMFORCE(type,buffer,length,datatype,id) \
        RECVASYNCNOMEM(type,buffer,length,datatype,id) 
#define SENDSYNCNOMEMFORCE(type,buffer,length,to,datatype) \
        SENDSYNCNOMEM(type,buffer,length,to,datatype)
#define RECVSYNCNOMEMFORCE(type,buffer,length,datatype)  \
        RECVSYNCNOMEM(type,buffer,length,datatype)
#define SENDWAITFORCE(type,buffer,length,to,datatype,id) 
#define SENDWAITNOMEMFORCE(type,buffer,length,to,datatype,id) 
#define RECVWAITFORCE(type,buffer,length,datatype,id) \
        RECVSYNC(type,buffer,length,datatype)
#define RECVWAITNOMEMFORCE(type,buffer,length,datatype,id)     \
        RECVWAITNOMEM(type,buffer,length,datatype,id)
#endif

#if NO_NATIVE_SUBSET
/* Here go standard routines when global operations are supported,
   but not operations on subsets.  This should define only those that
   have global versions available.  To use these,
   #define G....GLOB(...) for the operations that are available.
 */
#if defined(GISUMGLOB)
#define GISUM(val,n,work,procset)  {LOGOPSTART(LOG_GSUMS);if(!(procset)) \
			GISUMGLOB(val,n,work);else\
                        gisumset(val,n,work,procset);LOGOPEND(LOG_GSUME);}
#endif
#if defined(GDSUMGLOB)
#define GDSUM(val,n,work,procset)  {LOGOPSTART(LOG_GSUMS);if(!(procset)) \
			   GDSUMGLOB(val,n,work);else\
                           gdsumset(val,n,work,procset);LOGOPEND(LOG_GSUME);}
#endif
#if defined(GFSUMGLOB)
#define GFSUM(val,n,work,procset)  {LOGOPSTART(LOG_GSUMS);if(!(procset)) \
			   GFSUMGLOB(val,n,work);else\
                           gfsumset(val,n,work,procset);LOGOPEND(LOG_GSUME);}
#endif
#if defined(GIMAXGLOB)
#define GIMAX(val,n,work,procset)  {LOGOPSTART(LOG_GMAXS);if(!(procset)) \
			  GIMAXGLOB(val,n,work);else\
			  gihighset(val,n,work,procset);LOGOPEND(LOG_GMAXE);}
#endif
#if defined(GIMINGLOB)
#define GIMIN(val,n,work,procset)  {LOGOPSTART(LOG_GMINS);if(!(procset)) \
			    GIMINGLOB(val,n,work);else\
                            gilowset(val,n,work,procset);LOGOPEND(LOG_GMINE);}
#endif
#if defined(GDMAXGLOB)
#define GDMAX(val,n,work,procset)  {LOGOPSTART(LOG_GMAXS);if(!(procset)) \
			  GDMAXGLOB(val,n,work);else\
			  gdhighset(val,n,work,procset);LOGOPEND(LOG_GMAXE);}
#endif
#if defined(GDMINGLOB)
#define GDMIN(val,n,work,procset)  {LOGOPSTART(LOG_GMINS);if(!(procset)) \
			    GDMINGLOB(val,n,work);else\
                            gdlowset(val,n,work,procset);LOGOPEND(LOG_GMINE);}
#endif
#if defined(GIORGLOB)
#define GIOR(val,n,work,procset)   {if (!(procset)) \
                            GIORGLOB(val,n,work);else\
                             giorset( val, n, work, procset );}
#endif
#if defined(GSYNCGLOB)
#define GSYNC(procset)      {LOGSYNCSTART(LOG_GSYNCS);\
			     if(!(procset))GSYNCGLOB();\
                             else gsyncset(procset);LOGSYNCEND(LOG_GSYNCE);}
#endif
#if defined(GCOLGLOB)
#define GCOL(lbuf,lsize,gbuf,gsiz,glen,procset,datatype) {\
      LOGOPSTART(LOG_GCOLS);\
      if(!(procset))GCOLGLOB(lbuf,lsize,gbuf,gsiz,glen,datatype);else\
      gcolset(lbuf,lsize,gbuf,gsiz,glen,procset,datatype);LOGOPEND(LOG_GCOLE);}
#endif
#if defined(GCOLXGLOB)
#define GCOLX(lbuf,gsizes,gbuf,procset,datatype) {LOGOPSTART(LOG_GCOLS);\
        if(!(procset))GCOLXGLOB(lbuf,gsizes,gbuf,datatype);else\
        gcolxset(lbuf,gsizes,gbuf,procset,datatype);LOGOPEND(LOG_GCOLE);}
#endif
/* #define GTOKEN(procset,i) gtokenset(procset,i)*/
#if defined(GSCATTERGLOB)
#define GSCATTER(buf,size,issrc,procset,datatype) {LOGOPSTART(LOG_GSCATS);\
        if (!(procset)) {GSCATTERGLOB(buf,size,issrc,datatype);}else\
        gscatterset(buf,size,issrc,procset,datatype);LOGOPEND(LOG_GSCATE);}
#endif
#if defined(GSCATTERSRCGLOB)
#define GSCATTERSRC(buf,size,src,procset,datatype) {LOGOPSTART(LOG_GSCATS);\
        if (!(procset)){GSCATTERSRCGLOB(buf,size,src,datatype);}else\
        gscattersetsrc(buf,size,src,procset,datatype);LOGOPEND(LOG_GSCATE);}
#endif

#endif /* NO_NATIVE_SUBSET */

#ifdef NO_NATIVE_GLOBAL
extern void gcsumset(), gisumset(), gdsumset(), gfsumset();
extern void gchighset(), gihighset(), gfhighset(), gdhighset(),
            gclowset(),  gilowset(),  gflowset(),  gdlowset();
extern void gcorset(), giorset(), gcandset(), giandset();
extern void gopset(), gsyncset(),
	    gcolset(), gcolxset();
#if !defined(GCSUM)
#define GCSUM(val,n,work,procset)  {LOGOPSTART(LOG_GSUMS);\
			  gcsumset(val,n,work,procset);LOGOPEND(LOG_GSUME);}
#endif
#if !defined(GISUM)
#define GISUM(val,n,work,procset)  {LOGOPSTART(LOG_GSUMS);\
			  gisumset(val,n,work,procset);LOGOPEND(LOG_GSUME);}
#endif
#if !defined(GDSUM)
#define GDSUM(val,n,work,procset)  {LOGOPSTART(LOG_GSUMS);\
			  gdsumset(val,n,work,procset);LOGOPEND(LOG_GSUME);}
#endif
#if !defined(GFSUM)
#define GFSUM(val,n,work,procset)  {LOGOPSTART(LOG_GSUMS);\
			  gfsumset(val,n,work,procset);LOGOPEND(LOG_GSUME);}
#endif
#if !defined(GCMAX)
#define GCMAX(val,n,work,procset)  {LOGOPSTART(LOG_GMAXS);\
			  gchighset(val,n,work,procset);LOGOPEND(LOG_GMAXE);}
#endif
#if !defined(GIMAX)
#define GIMAX(val,n,work,procset)  {LOGOPSTART(LOG_GMAXS);\
			  gihighset(val,n,work,procset);LOGOPEND(LOG_GMAXE);}
#endif
#if !defined(GDMAX)
#define GDMAX(val,n,work,procset)  {LOGOPSTART(LOG_GMAXS);\
			  gdhighset(val,n,work,procset);LOGOPEND(LOG_GMAXE);}
#endif
#if !defined(GFMAX)
#define GFMAX(val,n,work,procset)  {LOGOPSTART(LOG_GMAXS);\
			  gfhighset(val,n,work,procset);LOGOPEND(LOG_GMAXE);}
#endif
#if !defined(GCMIN)
#define GCMIN(val,n,work,procset)  {LOGOPSTART(LOG_GMINS);\
			  gclowset(val,n,work,procset);LOGOPEND(LOG_GMINE);}
#endif
#if !defined(GIMIN)
#define GIMIN(val,n,work,procset)  {LOGOPSTART(LOG_GMINS);\
			  gilowset(val,n,work,procset);LOGOPEND(LOG_GMINE);}
#endif
#if !defined(GDMIN)
#define GDMIN(val,n,work,procset)  {LOGOPSTART(LOG_GMINS);\
			  gdlowset(val,n,work,procset);LOGOPEND(LOG_GMINE);}
#endif
#if !defined(GFMIN)
#define GFMIN(val,n,work,procset)  {LOGOPSTART(LOG_GMINS);\
			  gflowset(val,n,work,procset);LOGOPEND(LOG_GMINE);}
#endif

#if !defined(GCOR)
#define GCOR(val,n,work,procset)   \
    gcorset( val, n, work, procset )
#endif
#if !defined(GIOR)
#define GIOR(val,n,work,procset)   \
    giorset( val, n, work, procset )
#endif

#if !defined(GCAND)
#define GCAND(val,n,work,procset)   \
    gcandset( val, n, work, procset )
#endif
#if !defined(GIAND)
#define GIAND(val,n,work,procset)   \
    giandset( val, n, work, procset )
#endif

#if !defined(GSYNC)
#define GSYNC(procset)             {LOGSYNCSTART(LOG_GSYNCS);\
				gsyncset(procset);LOGSYNCEND(LOG_GSYNCE);}
#endif
#if !defined(GCOL)
#define GCOL(lbuf,lsize,gbuf,gsiz,glen,procset,datatype){\
      LOGOPSTART(LOG_GCOLS);\
      gcolset(lbuf,lsize,gbuf,gsiz,glen,procset,datatype);LOGOPEND(LOG_GCOLE);}
#endif
#if !defined(GCOLX)
#define GCOLX(lbuf,gsizes,gbuf,procset,datatype) {LOGOPSTART(LOG_GCOLS);\
        gcolxset(lbuf,gsizes,gbuf,procset,datatype);LOGOPEND(LOG_GCOLE);}
#endif
#if !defined(GTOKEN)
#define GTOKEN(procset,i) gtokenset(procset,i)
#endif
#if !defined(GSCATTER)
#define GSCATTER(buf,siz,issrc,procset,datatype) {LOGOPSTART(LOG_GSCATS);\
        gscatterset(buf,siz,issrc,procset,datatype);LOGOPEND(LOG_GSCATE);}
#endif
#if !defined(GSCATTERSRC)
#define GSCATTERSRC(buf,siz,src,procset,datatype) {LOGOPSTART(LOG_GSCATS);\
        gscattersetsrc(buf,siz,src,procset,datatype);LOGOPEND(LOG_GSCATE);}
#endif

#endif  /* NO_NATIVE_GLOBAL */

/* Finally, include the "forward" compatibility routines */
#include "comm/newnames.h"

/* At this point, PImytid is defined */
#ifdef DEBUG_TRACEBACK
/* REPLACE the definitions of DEBUGTBF to include the process id */
#undef DEBUGTBF
#undef DEBUGTBCF
#define DEBUGTBF(line,file) fprintf( stderr, "[%d] Line %d in %s\n", PImytid, line, file )
#define DEBUGTBCF(line,file,c) \
    fprintf( stderr, "[%d] Line %d in %s: %s\n", PImytid, line, file, c )
#endif

/* The man pages for the macros are in the file "readme", since any changes to 
   this file can force recompilation of all parallel code. */

#endif

