#include <stdio.h>
#define EUI_GLOBAL_AVAILABLE
#define DEFAULT_GREPS 50

#include "comm/examples/angst/angst.h"
#include "testing/lrctx.h"

/*
    This is a simple program to test the performance of some collective
    operations.

    I've re-worked this to be more complete and to use the LR routines
    to obtain reproducible times .

    To do this, LRRunSingleTest requires that the tested function have the
    form:

    double = f( double, void * )

    where double is the time returned.

 */     

/* Structure for the test routines */
typedef struct {
    ProcSet *pset;       /* Procset to test over */
    int     src;         /* Source (for scatter) */
    int     lreps;       /* local repititions */
    } GOPctx;

#define GETMEM(type,nitem,p) {p = (type *)MALLOC( nitem * sizeof(type) );\
				  CHKPTRN(p);}
#define SETMEM(nitem,p) {int _i; for (_i=0; _i<nitem; _i++) p[_i] = 0; }

/* Eventually, include :
   
     datatype (double, float, int, char)
     operation (sum, min, max, and, or, collect, scatter)
   
   and accept any combination.
 */
typedef enum { GDOUBLE, GFLOAT, GINT, GCHAR } GDatatype;
typedef enum { GOPSUM, GOPMIN, GOPMAX, GOPSYNC, GOPBCAST, GOPCOL } GOperation;

/* Test routines.
   These are really all almost the same; just different enough that we do them
   with pick and put programming rather than a single routine or macro.
   A feature of the LR routines is that ALL tested routines must return the
   same time, because this time is used to adaptively decide when to quit.
 */

/* First are the Chameleon versions */
double TestGDSum( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int    len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
double *lval, *work, time;
SYusc_time_t t0, t1;

GETMEM(double,len,lval);
GETMEM(double,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgdsum( lval, len, work, pset );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}

double TestGISum( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
int     *lval, *work;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
GETMEM(int,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgisum( lval, len, work, pset );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}

double TestGScat( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
int     *lval;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIbcast( lval, len, PSISROOT(pset), pset, MSG_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
return time;
}

double TestGCol( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps, gsize, glen;
ProcSet *pset = ctx->pset;
int     *lval, *gval;
double  time;
SYusc_time_t t0, t1;

gsize = len * PSnumtids(pset);
GETMEM(int,len,lval);
GETMEM(int,gsize,gval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgcol( lval, len, gval, gsize, &glen, pset, MSG_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( gval );
return time;
}

double TestGSync( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
double  time;
SYusc_time_t t0, t1;

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgsync( pset );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
return time;
}

/* Next are the "vendor" versions (when available) */
#ifdef GDSUMGLOB
double TestGDSumGlob( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int    len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
double *lval, *work, time;
SYusc_time_t t0, t1;

GETMEM(double,len,lval);
GETMEM(double,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GDSUMGLOB( lval, len, work );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}
#endif

#ifdef GISUMGLOB
double TestGISumGlob( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
int     *lval, *work;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
GETMEM(int,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GISUMGLOB( lval, len, work );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}
#endif

#ifdef GSCATTERGLOB
double TestGScatGlob( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
int     *lval;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GSCATTERGLOB( lval, len, PSISROOT(pset), MSG_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
return time;
}
#endif

#ifdef GCOLGLOB
double TestGColGlob( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps, gsize, glen;
ProcSet *pset = ctx->pset;
int     *lval, *gval;
double  time;
SYusc_time_t t0, t1;

gsize = len * PSnumtids(pset);
GETMEM(int,len,lval);
GETMEM(int,gsize,gval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GCOLGLOB( lval, len, gval, gsize, &glen, G_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( gval );
return time;
}
#endif 

#ifdef GSYNCGLOB
double TestGSyncGlob( dlen, ctx )
double dlen;
GOPctx *ctx;
{
int     len = (int)dlen, i, reps = ctx->lreps;
ProcSet *pset = ctx->pset;
double  time;
SYusc_time_t t0, t1;

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GSYNCGLOB();
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 ) / (double)(reps);
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
return time;
}
#endif 

extern void gsetopL(), gsetopT(), gsetopTP();
extern void gscattersetR();

int worker();

static int use_native = 0;
static ProcSet *pset  = PSAllProcs;

/* Stub for general parallel call */
main(argc,argv)
int argc;
char *argv[];
{
return PICall( worker, argc, argv );
}

int worker(argc,argv)
int argc;
char *argv[];
{
int c;
double (* f)();
long   len, error_flag;
double t;
long   first,last,incr, svals[3];
long   pvals[3], p;
int    rrsize;
char   psetname[50];
int    psetlen = 50;
GOPctx ctx;
GOperation op    = GOPSYNC;
GDatatype  dtype = GDOUBLE;

SYusc_init();

if (SYArgHasName( &argc, argv, 1, "-help" )) {
  if (PImytid == 0) PrintHelp( argv );
  return 0;
  }
  
/* Collective operations take longer than single operations, so
   choose a smaller repitition value. Eventually, we should copy the
   twin code that dynamically chooses the lreps value. */
ctx.lreps  = DEFAULT_GREPS;
ctx.pset   = 0;
ctx.src    = 0;
svals[0]   = sizeof(double);
svals[1]   = 128 * sizeof(double);     /* 1024 usually */
svals[2]   =   4 * sizeof(double);     /* 32 usually */
pvals[0]   = 1024*1024;
pvals[1]   = pvals[0];
pvals[2]   = pvals[0];
use_native = 0;

if (SYArgGetString( &argc, argv, 1, "-pset", psetname, 50 )) {
    int n1, m1, i;
    ctx.pset = PSCreate( 1 );
    sscanf( psetname, "%d-%d", &n1, &m1 );
    for (i=n1; i<=m1; i++) {
	PSAddMember( ctx.pset, &i, 1 );
	}
    PSCompile( ctx.pset );
    }
if (SYArgGetInt( &argc, argv, 1, "-rr", &rrsize )) {
    PISetRRSize( rrsize );
    }
if (SYArgHasName( &argc, argv, 1, "-ring" )) {
    if (PImytid == 0)
	fprintf( stderr, "Ring not yet supported\n" );
    return 0;
    /*    PISetCombFunc( gsetopR ); */
    }
if (SYArgHasName( &argc, argv, 1, "-line" )) {
    PISetCombFunc( gsetopL );
    PISetScatterFunc( gscattersetR, (void (*)())0 );
    }
if (SYArgHasName( &argc, argv, 1, "-fast" )) {
    PISetNbrRoutines( PIDefFastTree, (void (*)())0, (void (*)())0 );
    PISetCollectiveTree( PISetTreeNodesFast );
    PISetupCollectiveTree();
    }
if (SYArgHasName( &argc, argv, 1, "-tree" ))
    PISetCombFunc( gsetopT );

/* Choose the operations */
if (SYArgHasName( &argc, argv, 1, "-dsum" )) {
    op    = GOPSUM;
    dtype = GDOUBLE;
    }
if (SYArgHasName( &argc, argv, 1, "-isum" )) {
    op    = GOPSUM;
    dtype = GINT;
    }
if (SYArgHasName( &argc, argv, 1, "-sync" )) {
    op    = GOPSYNC;
    }
if (SYArgHasName( &argc, argv, 1, "-scatter" )) {
    op    = GOPBCAST;
    dtype = GINT;
    }
if (SYArgHasName( &argc, argv, 1, "-col" )) {
    op    = GOPCOL;
    dtype = GINT;
    }
if (SYArgHasName( &argc, argv, 1, "-native" )) {
    use_native = 1;
    }

/* Convert operation and dtype to routine */
f = 0;
switch (op) {
    case GOPSUM:
        switch (dtype) {
	    case GDOUBLE: 
	    f = TestGDSum;
#if defined(GDSUMGLOB)
	    if (use_native) f = TestGDSumGlob;
#else 
	    if (use_native) f = 0;
#endif
	    break;
	    case GINT:
	    f = TestGISum;
#if defined(GISUMGLOB)
	    if (use_native) f = TestGISumGlob;
#else 
	    if (use_native) f = 0;
#endif
	    break;
	    }
	break;
    case GOPMIN:
    case GOPMAX:
	f = 0;
	break;
    case GOPCOL:
	f = TestGCol;
#if defined(GCOLGLOB)
	if (use_native) f = TestGColGlob;
#else 
	if (use_native) f = 0;
#endif
	break;
    case GOPBCAST:
	f = TestGScat;
#if defined(GSCATTERGLOB)
	if (use_native) f = TestGScatGlob;
#else 
	if (use_native) f = 0;
#endif	
	break;
    case GOPSYNC:
	f = TestGSync;
#if defined(GSYNCGLOB)
	if (use_native) f = TestGSyncGlob;
#else 
	if (use_native) f = 0;
#endif
	break;
    }

if (!f) {
    if (PImytid == 0)
	fprintf( stderr, "Operation and datatype not supported\n" );
    SYexitall( (char *)0, 2 );
    }

if (SYArgGetIntVec( &argc, argv, 1, "-pkt", 3, pvals )) {
    /* Only pipelined version for now */
    PISetCombFunc( gsetopTP );
    /* Native version does not have packet size control */
    if (use_native) pvals[1] = pvals[0];
    }
SYArgGetIntVec( &argc, argv, 1, "-size", 3, svals );
SYArgGetInt(    &argc, argv, 1, "-reps", &ctx.lreps );

/* sync only makes sense for a single size */
if (op == GOPSYNC) svals[0] = svals[1] = 0;

first = svals[0];
last  = svals[1];
incr  = svals[2];

for (p=pvals[0]; p<=pvals[1]; p += pvals[2]) {
    PISetPacketSize( p );	
    time_function( &ctx, f, first,last,incr);
    }

PIgsync(PSAllProcs);
if (PImytid == 0)
    printf( "Done\n" );
return 0;    
}

time_function( fctx, f, first, last, incr )
GOPctx *fctx;
double (* f)();
int    first, last, incr;
{
LRctx *lrctx;
long len,myproc;
double mean_time,rate;
double t;
double s, r;

myproc = PImytid;
if (myproc == 0) {
    fprintf( stdout, "\n#len\ttime\t\tave time (us)\trate\n"); 
    fflush( stdout );
    }
lrctx = LRCreate();
for(len=first;len<=last;len+=incr){
    t         = LRRunSingleTest( lrctx, f, fctx, (double)(len) );
    mean_time = ( t ) * 1.0e6;  /* convert to us */
    rate      = (double)(len)/(mean_time*(1e-6));
    if(myproc == 0) {
	fprintf( stdout, "%d\t%f\t%f\t%f\n", len, t, mean_time, rate);
	fflush( stdout );
	}
    }
LRComputeParams( lrctx, &s, &r );
LRDestroy( lrctx );
if (myproc == 0) 
    fprintf( stdout, "# linear regression model is (%e + %e * len)\n", s, r );
}
 

PrintHelp( argv )
char **argv;
{
  fprintf( stderr, "%s - test collection operations for speed\n", argv[0] );
  fprintf( stderr, 
 "-rr n -ring -line -tree -dsum -isum -sync -scatter -pkt len stride cnt\n" );
  fprintf( stderr, "-native -size len stride cnt -reps n -fast\n" );

  fprintf( stderr, "\nTests:\n" );
  fprintf( stderr, "-dsum     : reduction (double precision)\n" );
  fprintf( stderr, "-isum     : reduction (integer)\n" );
  fprintf( stderr, "-sync     : synchronization\n" );
  fprintf( stderr, "-scatter  : scatter\n" );

  fprintf( stderr, "\nTest control:\n" );
  fprintf( stderr, "-size len stride cnt : size of messages\n" );
  fprintf( stderr, "-reps n              : number of times to repeat test\n" );
  fprintf( stderr, 
	  "-pset n-m            : processor set consisting of nodes n to m" );

  fprintf( stderr, "\nCollective communication algorithms:\n" );
  fprintf( stderr, 
   "-rr n   : use a ready-receiver version for messages n bytes or longer\n" );
  fprintf( stderr, 
   "          (many operations do not yet have rr versions)\n" );
  fprintf( stderr, "-ring   : use a ring algorithm\n" );
  fprintf( stderr, "-tree   : use a tree algorithm\n" );
  fprintf( stderr, "-line   : use a linear algorithm\n" );
  fprintf( stderr, "-native : use the native routines (if available)\n" );

  fprintf( stderr, "\nOptions for algorithms\n" );
  fprintf( stderr, "-pkt len stride cnt : use packets of length len\n" );
  fprintf( stderr, 
         "-fast               : use a (possibly) contention-free tree\n" );
}
